@@ -44,6 +44,10 @@ enum {
 #define COLOR_WHITE   0x00ffffff
 #define CHECK_SIZE    50
 
+#define RECALC_BOUNDS(image, item) \
+get_bounds ((image), &(item)->x1, &(item)->y1, &(item)->x2, &(item)->y2)
+
+
 static void gnome_canvas_thumb_class_init    (GnomeCanvasThumbClass *class);
 static void gnome_canvas_thumb_init          (GnomeCanvasThumb      *image);
 static void gnome_canvas_thumb_destroy       (GtkObject             *object);
@@ -137,7 +141,7 @@ gnome_canvas_thumb_class_init (GnomeCanvasThumbClass *class)
                  PROP_PIXBUF,
                  g_param_spec_object ("pixbuf", NULL, NULL,
                                       GDK_TYPE_PIXBUF,
-                                      (G_PARAM_READABLE | G_PARAM_WRITABLE)));
+                                      G_PARAM_WRITABLE));
 	g_object_class_install_property
                 (gobject_class,
                  PROP_X,
@@ -178,7 +182,6 @@ gnome_canvas_thumb_class_init (GnomeCanvasThumbClass *class)
 static void
 gnome_canvas_thumb_init (GnomeCanvasThumb *image)
 {
-	image->pixbuf = NULL;
 	image->x = 0.0;
 	image->y = 0.0;
 	image->width = DEFAULT_WIDTH;
@@ -190,12 +193,12 @@ gnome_canvas_thumb_init (GnomeCanvasThumb *image)
 static void
 free_pixmap_and_mask (GnomeCanvasThumb *image)
 {
-	if (image->pixmap) {
+	if (image->pixmap != NULL) {
 		g_object_unref (image->pixmap);
 		image->pixmap = NULL;
 	}
 
-	if (image->mask) {
+	if (image->mask != NULL) {
 		g_object_unref (image->mask);
 		image->mask = NULL;
 	}
@@ -205,6 +208,91 @@ free_pixmap_and_mask (GnomeCanvasThumb *image)
 }
 
 
+static void
+render_to_pixmap (GnomeCanvasThumb *image,
+		  GdkPixbuf        *pixbuf)
+{
+	static GdkPixbuf *last_pixbuf = NULL;
+	static GdkPixmap *last_pixmap = NULL;
+	static GdkBitmap *last_mask   = NULL;
+	GdkPixbuf        *tmp;
+
+	if (pixbuf != NULL) {
+		image->iwidth = gdk_pixbuf_get_width (pixbuf);
+		image->iheight = gdk_pixbuf_get_height (pixbuf);
+	}
+
+	free_pixmap_and_mask (image);
+
+	if ((pixbuf != NULL) && (last_pixbuf == pixbuf)) {
+		if (last_pixmap != NULL)
+			image->pixmap = g_object_ref (last_pixmap);
+		else
+			image->pixmap = NULL;
+
+		if (last_mask != NULL)
+			image->mask = g_object_ref (last_mask); 
+		else
+			image->mask = NULL;
+
+		return;
+	} 
+
+	if (last_pixbuf != NULL) {
+		g_object_unref (last_pixbuf);
+		last_pixbuf = NULL;
+	}
+
+	if (last_pixmap != NULL) {
+		g_object_unref (last_pixmap);
+		last_pixmap = NULL;
+	}
+
+	if (last_mask != NULL) {
+		g_object_unref (last_mask);
+		last_mask = NULL;
+	}
+
+	if (pixbuf == NULL) {
+		image->pixmap = NULL;
+		image->mask = NULL;
+		return;
+	}
+
+	last_pixbuf = g_object_ref (pixbuf);
+
+	if (gdk_pixbuf_get_has_alpha (pixbuf))
+		tmp = gdk_pixbuf_composite_color_simple (
+					 pixbuf, 
+					 image->iwidth,
+					 image->iheight, 
+					 GDK_INTERP_NEAREST,
+					 255,
+					 CHECK_SIZE, 
+					 COLOR_WHITE, 
+					 COLOR_WHITE);
+	else 
+		tmp = g_object_ref (pixbuf);
+
+	gdk_pixbuf_render_pixmap_and_mask (tmp,
+					   &(last_pixmap), 
+					   &(last_mask), 
+					   112);
+	
+	if (last_pixmap != NULL)
+		image->pixmap = g_object_ref (last_pixmap);
+	else
+		image->pixmap = NULL;
+	
+	if (last_mask != NULL)
+		image->mask = g_object_ref (last_mask); 
+	else
+		image->mask = NULL;
+
+	g_object_unref (tmp);
+}
+
+
 static void
 gnome_canvas_thumb_destroy (GtkObject *object)
 {
@@ -217,51 +305,14 @@ gnome_canvas_thumb_destroy (GtkObject *object)
 
 	free_pixmap_and_mask (image);
 
-	if (image->pixbuf != NULL) {
-		g_object_unref (G_OBJECT (image->pixbuf));
-		image->pixbuf = NULL;
-	}
+	/* do a fake rendering to free cache data. */
+	render_to_pixmap (image, NULL);
 
 	if (GTK_OBJECT_CLASS (parent_class)->destroy)
 		GTK_OBJECT_CLASS (parent_class)->destroy (object);
 }
 
 
-static void
-render_to_pixmap (GnomeCanvasThumb *image,
-		  GdkPixbuf *pixbuf)
-{
-	if (image->pixbuf != NULL) {
-		g_object_unref (G_OBJECT (image->pixbuf));
-		image->pixbuf = NULL;
-	}
-
-	image->iwidth = gdk_pixbuf_get_width (pixbuf);
-	image->iheight = gdk_pixbuf_get_height (pixbuf);
-
-	if (gdk_pixbuf_get_has_alpha (pixbuf))
-		image->pixbuf = gdk_pixbuf_composite_color_simple (
-					   pixbuf, 
-					   image->iwidth,
-					   image->iheight, 
-					   GDK_INTERP_NEAREST,
-					   255,
-					   CHECK_SIZE, 
-					   COLOR_WHITE, 
-					   COLOR_WHITE);
-	else {
-		image->pixbuf = pixbuf;
-		g_object_ref (G_OBJECT (image->pixbuf));
-	}
-	
-	free_pixmap_and_mask (image);
-	gdk_pixbuf_render_pixmap_and_mask (image->pixbuf, 
-					   &(image->pixmap), 
-					   &(image->mask), 
-					   112);
-}
-
-
 /* Set_property handler for the pixbuf canvas item */
 static void
 gnome_canvas_thumb_set_property (GObject       *object,
@@ -269,13 +320,13 @@ gnome_canvas_thumb_set_property (GObject       *object,
 				 const GValue  *value,
 				 GParamSpec    *pspec)
 {
-	GnomeCanvasItem *item;
+	GnomeCanvasItem  *item;
 	GnomeCanvasThumb *image;
-	gint update;
-	GdkPixbuf *pixbuf;
+	int               update;
+	GdkPixbuf        *pixbuf;
 
-	item = GNOME_CANVAS_ITEM (object);
-	image = GNOME_CANVAS_THUMB (object);
+	item = (GnomeCanvasItem*) object;
+	image = (GnomeCanvasThumb*) object;
 
 	update = FALSE;
 
@@ -283,8 +334,6 @@ gnome_canvas_thumb_set_property (GObject       *object,
         case PROP_PIXBUF:
 		pixbuf = GDK_PIXBUF (g_value_get_object (value));
 		g_return_if_fail (pixbuf != NULL);
-		if (pixbuf == image->pixbuf) 
-			return;
 
 		render_to_pixmap (image, pixbuf);
 
@@ -328,15 +377,9 @@ gnome_canvas_thumb_get_property (GObject       *object,
 				 GValue        *value,
 				 GParamSpec    *pspec)
 {
-	GnomeCanvasThumb *image;
-
-	image = GNOME_CANVAS_THUMB (object);
+	GnomeCanvasThumb *image = (GnomeCanvasThumb*) object;
 
         switch (param_id) {
-        case PROP_PIXBUF:
-                g_value_set_object (value, G_OBJECT (image->pixbuf));
-                break;
-
         case PROP_X:
                 g_value_set_double (value, image->x);
                 break;
@@ -363,9 +406,7 @@ gnome_canvas_thumb_get_property (GObject       *object,
 static void
 gnome_canvas_thumb_realize (GnomeCanvasItem *item)
 {
-	GnomeCanvasThumb *image;
-
-	image = GNOME_CANVAS_THUMB (item);
+	GnomeCanvasThumb *image = (GnomeCanvasThumb*) item;
 
 	if (parent_class->realize)
 		(* parent_class->realize) (item);
@@ -378,9 +419,7 @@ gnome_canvas_thumb_realize (GnomeCanvasItem *item)
 static void
 gnome_canvas_thumb_unrealize (GnomeCanvasItem *item)
 {
-	GnomeCanvasThumb *image;
-
-	image = GNOME_CANVAS_THUMB (item);
+	GnomeCanvasThumb *image = (GnomeCanvasThumb*) item;
 
 	if (! item->canvas->aa) {
 		g_object_unref (image->gc);
@@ -397,11 +436,9 @@ get_bounds (GnomeCanvasThumb *image,
 	    double *px1, double *py1, 
 	    double *px2, double *py2)
 {
-        GnomeCanvasItem *item;
-	double i2c[6];
-	ArtDRect i_bbox, c_bbox;
-
-        item = GNOME_CANVAS_ITEM (image);
+        GnomeCanvasItem *item = (GnomeCanvasItem*) image;
+	double           i2c[6];
+	ArtDRect         i_bbox, c_bbox;
 
         i_bbox.x0 = image->x;
         i_bbox.y0 = image->y;
@@ -419,35 +456,14 @@ get_bounds (GnomeCanvasThumb *image,
 }
 
 
-/* Get's the image bounds expressed as item-relative coordinates. */
-static void
-get_bounds_item_relative (GnomeCanvasThumb *image, 
-			  double *px1, double *py1, 
-			  double *px2, double *py2)
-{
-	*px1 = image->x;
-	*py1 = image->y;
-	*px2 = image->x + image->width;
-	*py2 = image->y + image->height;
-}
-
-
-static void
-recalc_bounds (GnomeCanvasThumb *image)
-{
-	GnomeCanvasItem *item = GNOME_CANVAS_ITEM (image);
-	get_bounds (image, &item->x1, &item->y1, &item->x2, &item->y2);
-}
-
-
 static void
 gnome_canvas_thumb_update (GnomeCanvasItem *item, 
-			   double *affine, 
-			   ArtSVP *clip_path, 
-			   int flags)
+			   double          *affine, 
+			   ArtSVP          *clip_path, 
+			   int              flags)
 {
-	GnomeCanvasThumb *image = GNOME_CANVAS_THUMB (item);
-	ArtDRect i_bbox, c_bbox;
+	GnomeCanvasThumb *image = (GnomeCanvasThumb*) item;
+	ArtDRect          i_bbox, c_bbox;
 
 	if (parent_class->update)
 		(* parent_class->update) (item, affine, clip_path, flags);
@@ -457,25 +473,29 @@ gnome_canvas_thumb_update (GnomeCanvasItem *item,
 				     image->cy, 
 				     image->cx + image->cwidth + 1, 
 				     image->cy + image->cheight + 1);
-	
-	recalc_bounds (image);
-	get_bounds_item_relative (image, 
-				  &i_bbox.x0, &i_bbox.y0, 
-				  &i_bbox.x1, &i_bbox.y1);
+
+	RECALC_BOUNDS (image, item);
+
+	/* get the image bounds expressed as item-relative coordinates. */
+
+	i_bbox.x0 = image->x;
+	i_bbox.y0 = image->y;
+	i_bbox.x1 = image->x + image->width;
+	i_bbox.y1 = image->y + image->height;
+
+	/**/
+
 	art_drect_affine_transform (&c_bbox, &i_bbox, affine);
 	
-	image->cx = c_bbox.x0;
-	image->cy = c_bbox.y0;
-	image->cwidth = (int) (c_bbox.x1 - c_bbox.x0);
+	image->cx      = c_bbox.x0;
+	image->cy      = c_bbox.y0;
+	image->cwidth  = (int) (c_bbox.x1 - c_bbox.x0);
 	image->cheight = (int) (c_bbox.y1 - c_bbox.y0);
 	
 	c_bbox.x0--;
 	c_bbox.y0--;
-	c_bbox.x1++;
-	c_bbox.y1++;
-	
-	c_bbox.x1++;
-	c_bbox.y1++;
+	c_bbox.x1 += 2;
+	c_bbox.y1 += 2;
 	
 	gnome_canvas_request_redraw (item->canvas, c_bbox.x0, c_bbox.y0, c_bbox.x1, c_bbox.y1);
 }
@@ -484,11 +504,14 @@ gnome_canvas_thumb_update (GnomeCanvasItem *item,
 static void
 recalc_if_needed (GnomeCanvasThumb *image)
 {
+	GnomeCanvasItem *item = (GnomeCanvasItem*) image;
+
 	if (! image->need_recalc)
 		return;
 	image->need_recalc = FALSE;
 
-	recalc_bounds (image);
+	RECALC_BOUNDS (image, item);
+
 	if (image->gc)
 		gdk_gc_set_clip_mask (image->gc, image->mask);
 }
@@ -498,15 +521,13 @@ static void
 gnome_canvas_thumb_draw (GnomeCanvasItem *item, GdkDrawable *drawable,
 			  int x, int y, int width, int height)
 {
-	GnomeCanvasThumb *image;
+	GnomeCanvasThumb *image = (GnomeCanvasThumb*) item;
 	GtkStyle         *style;
         int               x1, y1;
 
-	image = GNOME_CANVAS_THUMB (item);
-
 	recalc_if_needed (image);
 
-	style = GTK_WIDGET (item->canvas)->style;
+	style = ((GtkWidget *) (item->canvas))->style;
 
 	x1 = image->cx - x;
 	y1 = image->cy - y;
@@ -567,12 +588,10 @@ gnome_canvas_thumb_point (GnomeCanvasItem *item,
 			   int cx, int cy, 
 			   GnomeCanvasItem **actual_item)
 {
-	GnomeCanvasThumb *image;
+	GnomeCanvasThumb *image = (GnomeCanvasThumb*) item;
 	double x1, y1, x2, y2;
 	double dx, dy;
 
-	image = GNOME_CANVAS_THUMB (item);
-
 	*actual_item = item;
 
 	recalc_if_needed (image);
@@ -610,9 +629,7 @@ gnome_canvas_thumb_bounds (GnomeCanvasItem *item,
 			    double *x1, double *y1, 
 			    double *x2, double *y2)
 {
-	GnomeCanvasThumb *image;
-
-	image = GNOME_CANVAS_THUMB (item);
+	GnomeCanvasThumb *image = (GnomeCanvasThumb*) item;
 
 	*x1 = image->x;
 	*y1 = image->y;
@@ -629,7 +646,7 @@ gnome_canvas_thumb_render (GnomeCanvasItem *item, GnomeCanvasBuf *buf)
 
 void
 gnome_canvas_thumb_select (GnomeCanvasThumb *image,
-			    gboolean select)
+			    gboolean         select)
 {
 	g_return_if_fail (image != NULL);
 	g_return_if_fail (GNOME_IS_CANVAS_THUMB (image));
